/*
 * This file is part of LibQtTracker project
 *
 * Copyright (C) 2009, Nokia
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */
/*
 * tracker.h
 *
 *  Created on: Jan 5, 2009
 *      Author: Iridian Kiiskinen <ext-iridian.kiiskinen at nokia.com>
 */

#ifndef QTTRACKER_BACKENDS_TRACKER_TRACKER_P_H_
#define QTTRACKER_BACKENDS_TRACKER_TRACKER_P_H_

#include <QDBusPendingCallWatcher>
#include "../../liveresource_p.h"
//#include "../../rdfunbound_p.h"
#include "common_p.h"
#include "signals_p.h"

namespace SopranoLive
{
	namespace BackEnds
	{
		namespace Tracker
		{
			enum { BatchSPARQLUpdate = RDFQuery::SPARQLUpdate + 1 };

			class TrackerResource
				: public ResourceOfService
			{
			public:
				TrackerResource
						(QSharedPointer<RDFServiceImplBase> const &service, Node const &node);
				~TrackerResource();

				mutable QString has_signaling_type;

				TrackerService *tracker() const;

				void removeInterface(QUrl const &/*interface_name*/) const;

				LiveNodeModelBasePtr implGetProperties(RDFDerivedSelect *columns
						, TripleModulePtr const &triple_module = TripleModulePtr()) const;
			};

			struct TrackerService
				: RDFServiceBasic
			{
				TrackerService();
				~TrackerService();

				QUrl serviceIri() const;

				static QSharedPointer<RDFGraphCache> getRDFGraphCache(bool persistent = true);
				static QSharedPointer<TrackerService> get();

				mutable org::freedesktop::Tracker1::Resources resources;
				QString stem;
				int id_counter, query_id_counter, update_id_counter, sync_id_counter;
				RDFServicePtr local_service_;


				typedef QMap<QString, ClassUpdateDispatcher *> QString2Dispatcher;

				bool waitForFinishedPendingCall
						( QDBusPendingCall &pending_call
						, QEventLoop::ProcessEventsFlags flags = QEventLoop::AllEvents) const;

					template<typename Reply_>
				bool waitForFinished(Reply_ &reply)
				{
					if(!reply.isFinished())
					{
						QVariantMap::const_iterator process_flags
								= attributes_.constFind("process_events_on_block");
						if(process_flags != attributes_.constEnd())
							return waitForFinishedPendingCall
									( reply, process_flags->value<QEventLoop::ProcessEventsFlags>());
						reply.waitForFinished();
					}
					return reply.isFinished();
				}

				QString2Dispatcher class2dispatcher;
				QString2Dispatcher property2dispatcher;

				ClassUpdateDispatcher *dispatcherFromClass(QString const &class_name);
				ClassUpdateDispatcher *dispatcherFromProperty(QString const &property_name);

				void loadTrackerPrefixes();

				QVector<QStringList> rawExecuteQueryOnGraph
						( QString const &query, int type, RDFGraphImplBase const &graph);
				void rawInsertTripleToGraph
						( QString const &subject, QString const &predicate, QString const &object
						, RDFGraphImplBase const &graph);
				void rawDeleteTripleFromGraph
						( QString const &subject, QString const &predicate, QString const &object
						, RDFGraphImplBase const &graph);
				void rawLoadOnGraph
						(QUrl const &file_uri, RDFGraphImplBase const &graph);

				QUrl createUniqueIri(QString stem_text);

				TrackerResource *newLiveResource(Node const &node = Node());

				Node rawNodeFromString(QString string, CppType type);
				QString rawStringFromNode(Node const &node);

				RDFStrategyFlags preprocessQuery(RDFStrategyFlags strategy);

				LiveNodes modelQuery
						( QString const &select, QStringList const &columns
						, RDFStrategyFlags strategy = RDFStrategy::DefaultStrategy
						, QSharedPointer<QObject> const &parent = QSharedPointer<QObject>()
						, TripleModulePtr const &triple_module = TripleModulePtr());
				LiveNodeModelBasePtr baseImplModelQuery
						( RDFDerivedSelect *select
						, RDFStrategyFlags strategy = RDFStrategy::DefaultStrategy
						, QSharedPointer<QObject> const &parent = QSharedPointer<QObject>()
						, TripleModulePtr const &triple_module = TripleModulePtr());
				QList<LiveNodes> executeQuery
						( RDFQuery const &query
						, RDFStrategyFlags strategy = RDFStrategy::DefaultStrategy
						, QSharedPointer<QObject> const &parent = QSharedPointer<QObject>());

				void addRemoveResourceClause(RDFStatementList *deleter, RDFVariable const &res);
/*
				ClassUpdateDispatcher *classUpdateDispatcher
						( QUrl const &class_uri, QString const &ontology_prefix
						, QString const &class_name);
				ClassUpdateDispatcher *classUpdateDispatcher(QUrl const &class_uri);
				ClassUpdateDispatcher *classUpdateDispatcher
						( QString const &ontology_prefix, QString const &class_name);
*/
				void setVerbosity(unsigned verbosity);
				void setConsoleVerbosity(unsigned console_verbosity);
				void setLogVerbosity(unsigned log_verbosity);
				void setLogDir(QDir const &path);
				QVariant serviceAttribute(QString const &attribute_name) const;
				QVariantMap serviceAttributes() const;

				int defaultTimeout() const;
				void setDefaultTimeout(int timeout = -1);

				QStringList inferNotifyTypes(RDFVariable const &resource) const;

				TripleUpdateContext::Mode registerForTripleUpdates
						(SharedTripleUpdateContext const &context);
			};

			struct TrackerContentModule
				: ContentModule_Simple
			{
				QSharedPointer<TrackerService> service_;
				RDFDerivedSelect columns_;
				int query_id_;
				bool flush_on_attach_;

				// completion_watcher_ signals when the whole process is done
				// data_watcher_ will contain the reply of the data. Distinction made
				// to facilitate Sync as completion marker for update calls
				QSharedPointer<QDBusPendingCallWatcher> completion_watcher_, data_watcher_;

				TrackerContentModule
						( QSharedPointer<TrackerService> service, RDFDerivedSelect const &columns
						  , bool flush_on_attach = true
						  , ContentModule::ChangePolicy change_policy = ContentModule::ReportChanges);
				TrackerContentModule(TrackerContentModule const &cp);
				~TrackerContentModule();

				virtual QString getQueryText() const;
				void attached(RowStoreModel &model);
				void detached(RowStoreModel &model);

				ReadyFlags refreshModel_chain
						( LiveNodeModel::RefreshMode mode, InterfaceChain<ModelCommons> chain);

				bool alterStrategy_chain
						( RDFStrategyFlags flags, RDFStrategyFlags mask, QModelIndex const &index
						, InterfaceChain<ModelCommons> chain);

				ContentModulePtr copy() const;
			};

			// Responsibility: Executing a RDFUpdate with the Tracker backend;
			// getting informed when the update is done.

			// FIXME: a more natural relationship between these 2 than inheritance?
			struct TrackerUpdateModule
				: QObject, TrackerContentModule
			{
				Q_OBJECT

			public:
				RDFUpdate update_;
				bool batch_mode_, sync_on_commit_;

				// Many TrackerUpdateModules may share the same D-Bus call.  This happens when the
				// tracker update query will contain multiple insert statements.  Each insert
				// statement gets its own model (and thus, a TrackerUpdateModule.) The
				// TrackerUpdateModule knows which is the index of the related insert statement in
				// the update, and also the other TrackerUpdateModules sharing the same D-Bus call.
				QList<TrackerUpdateModule*> sister_modules_;
				int my_index_;

				TrackerUpdateModule(QSharedPointer<TrackerService> service,
									RDFDerivedSelect const& columns,
									RDFUpdate const &update,
									bool flush_on_attach = true,
									bool batch_mode = false,
									bool sync_on_commit = false,
									ContentModule::ChangePolicy change_policy = ContentModule::ReportChanges);
				TrackerUpdateModule(TrackerUpdateModule const& cp);

				~TrackerUpdateModule();

				virtual QString getQueryText() const;

				ReadyFlags refreshModel_chain
						( LiveNodeModel::RefreshMode mode, InterfaceChain<ModelCommons> chain);


				void constructPendingCall();
				void setSisterModules(QList<TrackerUpdateModule*> const &sister_modules, int ix);

			public slots:
				void runSync();
			};

			struct TCM_TextQuery
				: TrackerContentModule
			{
				QString query_text_;

				static QSharedPointer<TCM_TextQuery> create
						( QSharedPointer<TrackerService> service, QString const &query_text
						, QStringList const &columns, bool flush_on_attach = true);
			private:
				friend class QSharedCreator<TCM_TextQuery>;
				TCM_TextQuery
						( QSharedPointer<TrackerService> service, QString const &query_text
						, RDFDerivedSelect const &columns, bool flush_on_attach);
				QString getQueryText() const;
				ContentModulePtr copy() const;
			};
		}
	}
}
#endif /* QTTRACKER_BACKENDS_TRACKER_TRACKER_P_H_ */
