/*
 * This file is part of libqttracker project
 *
 * Copyright (C) 2009, Nokia
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */
/*
 * updatermodule_p.h
 *
 *  Created on: Apr 7, 2010
 *      Author: "Iridian Kiiskinen"
 */

#ifndef SOPRANOLIVE_UPDATERMODULE_P_H_
#define SOPRANOLIVE_UPDATERMODULE_P_H_

#include "rdfcache_p.h"
#include "lnm_transitionstrategy_p.h"

namespace SopranoLive
{
	class UpdaterModule
		: public QObject
		, public ModuleOfDerivedModel<TripleModel>
		, public ModuleInterface<ModelCommons, TripleModel, UpdaterModule>
		, public ModuleInterface<TripleUpdateReceiver, TripleModel, UpdaterModule>
	{
		Q_OBJECT
	public:
		UpdaterModule(RDFStrategyFlags responsibility_modes);

		enum
		{
			OmnipresentResponsibilityModes = RDFStrategy::Running
		,	EndOfStreamSeen = RDFStrategyFlags(1) << RDFStrategy::offset_ContextSpecificFlags
		};

		RDFStrategyFlags responsibility_modes_;

		bool seenEndOfStream() const { return responsibility_modes_ & EndOfStreamSeen; }
		void setSeenEndOfStream(bool seen = true)
		{
			if(seen)
				responsibility_modes_ |= EndOfStreamSeen;
			else
				responsibility_modes_ &= ~RDFStrategyFlags(EndOfStreamSeen);
		}

		QList<RDFVariableLink> column_filters_;

		// contains a single entry for each pending update model.
		typedef QMap<LiveNodeModel const *, UpdateContextPtr> PendingUpdates;
		PendingUpdates pending_updates_;
		RDFSelect::OrderByInfo model_ordering_;

		// overrides of Module

		void attached(TripleModel &model);

		// overrides of ModelCommons

		bool alterStrategy_chain
				( RDFStrategyFlags flags, RDFStrategyFlags mask, QModelIndex const &index
				, InterfaceChain<ModelCommons> chain);

		ReadyFlags refreshModel_chain
				( LiveNodeModel::RefreshMode, InterfaceChain<ModelCommons> chain);

		// overrides of TripleUpdateReceiver

		bool tripleUpdates_chain
				( RawTriplesData const &triples, TripleUpdateContext const *context
				, InterfaceChain<TripleUpdateReceiver> chain);

		// local additions

		void changeState(RDFStrategyFlags flags, RDFStrategyFlags mask);

		RDFStrategyFlags constrainUpdateQuery
				( RawTriplesData const &triples, TripleUpdateContext const *context
				, RDFDerivedColumn *col, RDFDerivedSelect *update_select);

	public Q_SLOTS:
		void updateModelDerivedsUpdated();

	public:
		struct LiveUpdate
			: UpdateContext
		{
			LiveUpdate
					( RowStoreModelPtr const &model, RawTriplesData const &triples
					, SharedConstTripleUpdateContext const &update_context
					, RDFSelect::OrderByInfo const &model_ordering
					, SmoothTransitionModule::UpdateKeys const &update_keys);
			RowStoreModelPtr source_model_;
			RawTriplesData triples_;
			SharedConstTripleUpdateContext update_context_;
			RDFSelect::OrderByInfo model_ordering_;
			SmoothTransitionModule::UpdateKeys update_keys_;
			bool updateModel(LiveNodeModelBase *target) const;
		};

		struct ChainContext
			: TripleUpdateContext
		{
			ChainContext
					( QWeakPointer<TripleUpdateReceiver> const &receiver
					, RDFStatement const &triple_mask
					, QModelIndex const &index, RDFProperty const &link);
			RDFProperty link_;
			//! Contains information on aggregate chained property update dependency chains
		};
		typedef QExplicitlySharedDataPointer<ChainContext> SharedChainContext;

		SharedTripleUpdateContext createAndRegisterContext(RDFStatement const &statement, int col);
		RDFProperty recurseAndRegisterChainUpdates
				( RDFVariable const &source, RDFProperty const &link, int col
				, bool register_outermost = true);

		struct UpdateTransitionModule
			: ModuleOfDerivedModel<TripleModel>
			, ModuleInterface<ModelCommons, TripleModel, UpdaterModule>
		{
			bool updateTargetModule_chain
				( ModelModule *target_module, UpdateContext const &update_context
				, InterfaceChain<ModelCommons const> chain) const;
		};

		// streaming strategy handling functions

		void enableStreaming();
		void disableStreaming();

		bool initiateStreamPiece(bool first_in_stream);
		void receivedStreamPiece(int size);
		void cancelStreamPiece();
		void endOfStream(RDFStrategyFlags additional_resets = 0);

		struct StreamingUpdate
			: UpdateContext
		{
			RowStoreModel *source_;
			UpdaterModule *updater_;
			int last_requested_block_size_;
			StreamingUpdate(RowStoreModelPtr &source, UpdaterModule *updater, int requested_size)
				: UpdateContext(source), source_(source.data()), updater_(updater)
				, last_requested_block_size_(requested_size)
			{}
			bool updateModel(LiveNodeModelBase *target) const;
		};

		QSharedPointer<StreamingUpdate> streaming_update_;

		int streaming_block_size_, streaming_first_block_size_, streaming_remaining_limit_;
	};

}
#endif /* SOPRANOLIVE_UPDATERMODULE_P_H_ */
